#include "genesis.h"

#include "TranceS.h"
#include "VangelisS.h"

//#define USE_FRAMEBUFFER

#define WORK_W      BITMAP_WIDTH
#define WORK_H      BITMAP_HEIGHT
//#define WORK_W      24
//#define WORK_H      16
#define MAX_PARTIC  256
#define TAIL_LEN    6


typedef struct
{
    fix16 xpos;
    fix16 ypos;
} _particule_pos;

typedef struct
{
    fix16 xstep;
    fix16 ystep;
} _particule_mov;

typedef struct
{
    _particule_pos pos;
    _particule_mov mov;
} _particule;

_particule partics[MAX_PARTIC];
_particule_pos partics_pos[TAIL_LEN][MAX_PARTIC];
u16 numspartic[TAIL_LEN];

u16 numpartic;
s16 currentpartic;

fix16 baseposx;
fix16 baseposy;
fix16 gravity;
u8 vsync_on;
u8 paused;

static void handleInput();
static void joyEvent(u16 joy, u16 changed, u16 state);

static void initPartic(u16 num);
static void updatePartic(_particule_pos *part_pos, u16 num);
static void drawPartic(_particule_pos *part_pos, u16 num, u8 col);


int main()
{
    char col;

    JOY_setEventHandler(joyEvent);

    VDP_setHInterrupt(0);
    VDP_setHilightShadow(0);

    setBitmapMode(WORK_W, WORK_H);

    vsync_on = 1;
    paused = 0;
    col = 0xFF;

    /* Initialise particules */
    currentpartic = 0;
    baseposx = intToFix16(bitmap_height_pixel / 2);
    baseposy = intToFix16(bitmap_width_pixel / 2);
    gravity = FIX16(0.4);

    initPartic(64);

    /* Do main job here */
    while(1)
    {
        s16 prev;
        char str[8];

        handleInput();

        // calculates particules physic
        if (!paused)
        {
            numspartic[currentpartic] = numpartic;

            prev = currentpartic + 1;
            if (prev >= TAIL_LEN) prev = 0;

            updatePartic(partics_pos[currentpartic], numspartic[currentpartic]);

            // draw particules
            drawPartic(partics_pos[currentpartic], numspartic[currentpartic], col);
            drawPartic(partics_pos[prev], numspartic[prev], 0x0);
//            col += 0x11;

            // swap buffer
            currentpartic++;
            if (currentpartic >= TAIL_LEN) currentpartic = 0;

#ifdef USE_FRAMEBUFFER
            VDP_blitBitmap(0, 0, WORK_W, WORK_H);
#endif
        }

        // display frame rate
        showFPS();

        // display particul number
        intToStr(numpartic, str, 1);
        VDP_fillTileRect(APLAN, 0, 2, 3, 4, 1);
        VDP_drawText(APLAN, str, 0x8000, 2, 3);

        // display gravity
        fix16ToStr(gravity, str);
        VDP_fillTileRect(APLAN, 0, 2, 4, 5, 1);
        if (paused) VDP_drawText(APLAN, "PAUSE", 0x8000, 2, 4);
        else VDP_drawText(APLAN, str, 0x8000, 2, 4);

        if (vsync_on) VDP_waitVSync();
    }
}

static void handleInput()
{
    u16 value;

    value = JOY_readJoypad(JOY_1);

    if (value & BUTTON_A)
    {
        if (value & BUTTON_UP)
        {
            if (numpartic < MAX_PARTIC) numpartic++;
        }
        if (value & BUTTON_DOWN)
        {
            if (numpartic > 2) numpartic--;
        }
    }
    else if (value & BUTTON_B)
    {
        if (value & BUTTON_UP)
        {
            if (gravity > FIX16(0.1)) gravity -= FIX16(0.02);
        }
        if (value & BUTTON_DOWN)
        {
            if (gravity < FIX16(1.0)) gravity += FIX16(0.02);
        }
    }
    else
    {
        if (value & BUTTON_UP)
        {
            if (baseposy < intToFix16(bitmap_height_pixel - 10)) baseposy += FIX16(2.0);
        }
        if (value & BUTTON_DOWN)
        {
            if (baseposy > intToFix16(10)) baseposy -= FIX16(2.0);
        }
        if (value & BUTTON_LEFT)
        {
            if (baseposx > intToFix16(10)) baseposx -= FIX16(1.5);
        }
        if (value & BUTTON_RIGHT)
        {
            if (baseposx < intToFix16(bitmap_width_pixel - 10)) baseposx += FIX16(1.5);
        }
    }
}


static void joyEvent(u16 joy, u16 changed, u16 state)
{
    // START button state changed
    if (changed & BUTTON_START)
    {
        // START button pressed ?
        if (state & BUTTON_START)
        {
            // toggle pause
            if (paused) paused = 0;
            else paused = 1;
        }
    }

    if (changed & state & BUTTON_A)
    {
        if (isPlayingSample2C(AUDIO_CHANNEL1)) stopPlaySample2C(AUDIO_CHANNEL1);
        else startPlaySample2C(VangelisS, sizeof(VangelisS), AUDIO_CHANNEL1);
    }
    if (changed & state & BUTTON_B)
    {
//        if (isPlayingSample2C(AUDIO_CHANNEL2)) stopPlaySample2C(AUDIO_CHANNEL2);
//        else startPlaySample2C(tranceS, sizeof(tranceS), AUDIO_CHANNEL2);
        if (isPlayingSample2C(AUDIO_CHANNEL2)) stopPlaySample2C(AUDIO_CHANNEL2);
        else startPlaySample2C(tranceS, sizeof(tranceS), AUDIO_CHANNEL2);
    }

    // C button state changed
    if (changed & BUTTON_C)
    {
        // C button pressed ?
        if (state & BUTTON_C)
        {
            // toggle vsync
            if (vsync_on) vsync_on = 0;
            else vsync_on = 1;
        }
    }
}


static void initPartic(u16 num)
{
    _particule *p;
    _particule_pos *pp;
    s16 i;
    s16 j;

    numpartic = num;

    i = num;
    p = partics;
    while(i--)
    {
        p->pos.xpos = intToFix16(0);
        p->pos.ypos = intToFix16(0);
        p->mov.xstep = intToFix16(0);
        p->mov.ystep = intToFix16(0);
        p++;
    }

    j = TAIL_LEN;
    while(j--)
    {
        numspartic[j] = num;
        pp = partics_pos[j];
        i = num;
        while(i--)
        {
            pp->xpos = intToFix16(0);
            pp->ypos = intToFix16(0);
            pp++;
        }
    }
}

static void updatePartic(_particule_pos *part_pos, u16 num)
{
    _particule *p;
    _particule_pos *pp;
    s16 i;

    i = num;
    p = partics;
    pp = part_pos;
    while(i--)
    {
        if ((p->pos.xpos >= intToFix16(bitmap_width_pixel)) || (p->pos.xpos <= intToFix16(0)))
        {
            // re-init paticule
            p->pos.xpos = baseposx;
            p->pos.ypos = baseposy;
            p->mov.xstep = intToFix16(1) - (random() & (FIX16_FRAC_MASK << 1));
            p->mov.ystep = intToFix16(2) + (random() & (FIX16_FRAC_MASK << 3));
        }
        else if (p->pos.ypos <= intToFix16(0))
        {
            if (p->mov.ystep > -(gravity << 3))
            {
                // re-init paticule
                p->pos.xpos = baseposx;
                p->pos.ypos = baseposy;
                p->mov.xstep = intToFix16(1) - (random() & (FIX16_FRAC_MASK << 1));
                p->mov.ystep = intToFix16(2) + (random() & (FIX16_FRAC_MASK << 3));
            }
            else
            {
                // handle re-bound
                p->pos.xpos += p->mov.xstep;
                p->pos.ypos -= p->mov.ystep;
                p->mov.ystep = - p->mov.ystep;
                p->mov.ystep >>= 1;
            }
        }
        else
        {
            p->pos.xpos += p->mov.xstep;
            p->pos.ypos += p->mov.ystep;
            p->mov.ystep -= gravity;
        }

        pp->xpos = p->pos.xpos;
        pp->ypos = p->pos.ypos;

        p++;
        pp++;
    }
}

static void drawPartic(_particule_pos *part_pos, u16 num, u8 col)
{
    _particule_pos *p;
    s16 x, y;
    s16 i, maxy;

    i = num;
    maxy = bitmap_height_pixel;
    p = part_pos;
    while(i--)
    {
        x = fix16ToInt(p->xpos);
        y = maxy - fix16ToInt(p->ypos);

#ifdef USE_FRAMEBUFFER
        BITMAP_SETPIXEL_SAFE(x, y, col);
#else
        BITMAP_SETPIXELDIRECT_SAFE(x, y, col);
//            VDP_drawPixelDirect(x, y, col);
#endif

        p++;
    }
}
